unit clut;
//color lookup tables
interface
uses
{$IFDEF Unix} {$ELSE} Windows,{$ENDIF}
define_types, dialogs,Texture_3D_Unit,OpenGL1x, IniFiles, Prefs,StdCtrls,Classes;
type
  //TCLUT = array [0..255] of TGLRGBQuad;
  TCLUTnode =record
    intensity: byte;
    rgba: TGLRGBQuad;
  end;
  TCLUTnodeRA = array [0..255] of TCLUTnode;
  TCLUTrec =record
    numnodes: integer;
    min,max: single;
    nodes: TCLUTnodeRA;
  end;
var
  gCLUTrec: TCLUTrec;
procedure LinearMinMaxCLUT(lMin,lMax: TGLRGBQuad; var lCLUT: TLUT);
//creates lookup table for an arbitrary number of nodes
procedure GenerateLUT(lNodeRA: TCLUTrec;  var lCLUT: TLUT);
//set red,green,blue and alpha of a Quad
//set a node
function Node(lIntensity,lR,lG,lB,lA: byte): TCLUTnode;
//define a CLUT using text to define the nodes....
function CLUT2disk(lRead: boolean; lFilename: string; var lCLUTrec: TCLUTrec): boolean;
procedure CLUT2TLUT(lFilename: string; var lLUT: TLUT; var lCLUTrec: TCLUTrec);
procedure AutoContrast (var lCLUTrec: TCLUTrec);
procedure UpdateColorSchemes (var LUTdrop: TComboBox);
procedure SetItemNameX (lStr: string; var LUTdrop: TComboBox);
procedure rescale2RGB (var lTexture: TTexture; lMinScaled,lMaxScaled: single; var lCLUT: TLUT);
function CLUTDir: string;
procedure RangeRec (lMin,lMax: single);

implementation
uses SysUtils, mainunit, histogram;

procedure RangeRec (lMin,lMax: single);
begin
  gCLUTrec.min := lMin;
  gCLUTrec.max := lMax;
end;


function CLUTDir: string;
begin
  result := extractfilepath(paramstr(0))+'clut';
end;

procedure HideTopSlice (var lTexture: TTexture);
//we do not have a clipping plane for the top slice
//therefore, texture_3d_unit's "ClearSlice" sets the top (highest Z-dim) slice
//so all voxels have the minimum value
//however, if we choose an inverted color scheme we need to make the top slice transparent...
var
  lnVox,lVox,lI: integer;
begin
  lnVox := lTexture.Dim[1]*lTexture.Dim[2];
  lI := 3+ (4*lnVox * (lTexture.Dim[3]-1)); //A in RGBA quad indexed from zero
    for lVox := 1 to lnVox do begin
            lTexture.FiltImg^[lI] :=  0;//alpha
            inc(lI,4);
    end; //for each vox
end;

procedure rescale2RGB (var lTexture: TTexture; lMinScaled,lMaxScaled: single; var lCLUT: TLUT);
var
  lI,ValueI,lVox,lnVox: integer;
  lMin,lMax,lIntercept,lSlope: single;
  lCLUT8: TLUT;
  lRGBra: GLRGBQuadp0;
begin
  if lTexture.DataType <> GL_RGBA then
    exit;
  lnVox := lTexture.Dim[1]*lTexture.Dim[2]*lTexture.Dim[3]-1;
  if lnVox < 0 then
      exit;
  lMin := Scaled2Unscaled(lMinScaled,lTexture);
  lMax := Scaled2Unscaled(lMaxScaled,lTexture);
  if lMax = lMin then
    exit;
  if lMax < lMin then begin
      lSlope := lMax;
      lMax := lMin;
      lMin := lSlope;
  end;
  lIntercept := lMin;
  lSlope := 255/(lMax-lMin);
  lI := 0;
  if lTexture.RawUnscaledImg16 <> nil then begin //16
    for lVox := 0 to lnVox do begin
            ValueI := round((lTexture.RawUnscaledImg16^[lVox]-lINtercept)*lSlope);
            if ValueI < 0 then
              ValueI := 0;
            if ValueI > 255 then
              ValueI := 255;
            lTexture.FiltImg^[lI] :=  lCLUT[ValueI].rgbRed;//Luminance
            lTexture.FiltImg^[lI+1] := lCLUT[ValueI].rgbGreen;//Alpha
            lTexture.FiltImg^[lI+2] := lCLUT[ValueI].rgbBlue;//Alpha
            lTexture.FiltImg^[lI+3] := lCLUT[ValueI].rgbReserved;//Alpha
            inc(lI,4);
    end; //for each vox
  end else if lTexture.RawUnscaledImg32 <> nil then begin //32-bit
    for lVox := 0 to lnVox do begin
            ValueI := round((lTexture.RawUnscaledImg32^[lVox]-lIntercept)*lSlope);
            if ValueI < 0 then
              ValueI := 0;
            if ValueI > 255 then
              ValueI := 255;
            lTexture.FiltImg^[lI] :=  lCLUT[ValueI].rgbRed;//Luminance
            lTexture.FiltImg^[lI+1] := lCLUT[ValueI].rgbGreen;//Alpha
            lTexture.FiltImg^[lI+2] := lCLUT[ValueI].rgbBlue;//Alpha
            lTexture.FiltImg^[lI+3] := lCLUT[ValueI].rgbReserved;//Alpha
            inc(lI,4);
    end; //for each vox
  end else if (lTexture.RawUnscaledImg8 <> nil) then begin //8-bit
    //fast method - compute one calculation per index
    for lVox := 0 to 255 do begin
            ValueI := round((lVox-lIntercept)*lSlope);
            if ValueI < 0 then
              ValueI := 0;
            if ValueI > 255 then
              ValueI := 255;
            lCLUT8[lVox] :=  lCLUT[ValueI];//Luminance
    end; //for each vox
    lRGBra:= GLRGBQuadp0(lTexture.FiltImg);
    for lVox := 0 to lnVox do
            lRGBra^[lVox] := lCLUT8[lTexture.RawUnscaledImg8^[lVox]];
{//this works, ~30% slower than method above
  lI := 0;
    for lVox := 0 to lnVox do begin
            ValueI := lTexture.RawUnscaledImg8^[lVox];
            lTexture.FiltImg^[lI] :=  lCLUT8[ValueI].rgbRed;//Luminance
            lTexture.FiltImg^[lI+1] := lCLUT8[ValueI].rgbGreen;//Alpha
            lTexture.FiltImg^[lI+2] := lCLUT8[ValueI].rgbBlue;//Alpha
            lTexture.FiltImg^[lI+3] := lCLUT8[ValueI].rgbReserved;//Alpha
            inc(lI,4);
    end; //for each vox
//slow method - compute one calculation per voxel
    for lVox := 0 to lnVox do begin
            ValueI := round((lTexture.RawUnscaledImg8^[lVox]-lINtercept)*lSlope);
            if ValueI < 0 then
              ValueI := 0;
            if ValueI > 255 then
              ValueI := 255;
            lTexture.FiltImg^[lI] :=  lCLUT[ValueI].rgbRed;//Luminance
            lTexture.FiltImg^[lI+1] := lCLUT[ValueI].rgbGreen;//Alpha
            lTexture.FiltImg^[lI+2] := lCLUT[ValueI].rgbBlue;//Alpha
            lTexture.FiltImg^[lI+3] := lCLUT[ValueI].rgbReserved;//Alpha
            inc(lI,4);
    end; //for each vox}
  end;//8-bit
  if   (lMin < 0) and (lMax < 0) then
    HideTopSlice(lTexture);
  //EdgeDetect (lTexture,lMinScaled,lMaxScaled, lCLUT);

end;


(*procedure rescale2RGB (var lTexture: TTexture; lMin,lMax: single; var lCLUT: TCLUT);
begin

  if lTexture.DataType <> GL_RGBA then
    exit;
  if lTexture.HasUnscaled16 then
    rescale16toRGB (lTexture, round(lMin),round(lMax), lCLUT)
  else if lTexture.HasUnscaled32 then
    rescale32toRGB (lTexture, round(lMin),round(lMax), lCLUT)
  else


end;*)




function Node(lIntensity,lR,lG,lB,lA: byte): TCLUTnode;
begin
  result.intensity := lIntensity;
  result.rgba := RGBA(lR,lG,lB,lA);
end;

(*function IrcadLUT(lNodeRA: TCLUTnodeRA; lNodes,lMin,lMax: integer; var lCLUT: TCLUT):string;
//exports LUT in IrCad format to validate...
var
  lScale: single;
  lI: integer;
begin
    if lNodes < 1 then exit;
    result := '';
    lScale := (lMax-lMin)/255;
    for lI := 0 to (lNodes-1) do begin
      result := result+inttostr(lI+1);
      result := result+' '+ FloatToStrF(lNodeRA[lI].intensity*lScale +lMin, ffFixed, 8, 4);

      result := result+','+ FloatToStrF(lNodeRA[lI].rgba.rgbRed/255, ffFixed, 8, 4);
      result := result+','+ FloatToStrF(lNodeRA[lI].rgba.rgbGreen/255, ffFixed, 8, 4);
      result := result+','+ FloatToStrF(lNodeRA[lI].rgba.rgbBlue/255, ffFixed, 8, 4);
      result := result+','+ FloatToStrF(lNodeRA[lI].rgba.rgbReserved/255, ffFixed, 8, 4) ;

      result := result + ' | ';
    end;
end; *)

procedure GenerateLUT(lNodeRA: TCLUTrec; var lCLUT: TLUT);
var
  lSlope: single;
  lSpace,lI,lIprev,lS: integer;
  lMin,lMax: TCLUTnode;
begin
  //abba gCLUTrec := lNodeRA;
  if lNodeRA.numNodes < 2 then exit;
  lMin := lNodeRA.nodes[0];
  lMax := lNodeRA.nodes[lNodeRA.NumNodes-1];
  //check that nodes are in order...
  lIprev := lMin.intensity;
  for lI := 1 to (lNodeRA.numnodes-1) do begin
    if lNodeRA.nodes[lI].intensity <= lIprev then begin
      showmessage('Error, nodes not sorted or overlapping.');
      exit;
    end;
    lIprev := lNodeRA.nodes[lI].intensity;
  end;
  //clip values <= lMin to value of lMin
  for lI := 0 to lMin.Intensity do begin
    lCLUT[lI] := lMin.rgba;
  end;
  //clip values >= lMax to value of lMin
  for lI := lMax.Intensity to 255 do begin
    lCLUT[lI] := lMax.rgba;
  end;

  for lI := 0 to (lNodeRA.NumNodes-2) do begin
    lSpace := lNodeRA.nodes[lI+1].Intensity-lNodeRA.nodes[lI].Intensity;
    //interpolate red
    lSlope := (lNodeRA.nodes[lI+1].rgba.rgbRed-lNodeRA.nodes[lI].rgba.rgbRed)/lSpace;
    for lS := 1 to lSpace do
      lCLUT[lNodeRA.nodes[lI].Intensity+lS].rgbRed  :=lNodeRA.nodes[lI].rgba.rgbRed + round(lS * lSlope);
    //interpolate green
    lSlope := (lNodeRA.nodes[lI+1].rgba.rgbGreen-lNodeRA.nodes[lI].rgba.rgbGreen)/lSpace;
    for lS := 1 to lSpace do
      lCLUT[lNodeRA.nodes[lI].Intensity+lS].rgbGreen  :=lNodeRA.nodes[lI].rgba.rgbGreen + round(lS * lSlope);
    //interpolate blue
    lSlope := (lNodeRA.nodes[lI+1].rgba.rgbBlue-lNodeRA.nodes[lI].rgba.rgbBlue)/lSpace;
    for lS := 1 to lSpace do
      lCLUT[lNodeRA.nodes[lI].Intensity+lS].rgbBlue  :=lNodeRA.nodes[lI].rgba.rgbBlue + round(lS * lSlope);
    //interpolate alpha
    lSlope := (lNodeRA.nodes[lI+1].rgba.rgbreserved-lNodeRA.nodes[lI].rgba.rgbreserved)/lSpace;
    for lS := 1 to lSpace do
      lCLUT[lNodeRA.nodes[lI].Intensity+lS].rgbreserved  :=lNodeRA.nodes[lI].rgba.rgbreserved + round(lS * lSlope);

    (*for lS := 1 to lSpace do
      lNodeRA[lI+lS].rgba.rgbRed := lNodeRA[lI].rgba.rgbRed + round(lS * lSlope);
    //interpolate green
    lSlope := (lNodeRA[lI+1].rgba.rgbGreen-lNodeRA[lI].rgba.rgbGreen)/lSpace;
    for lS := 1 to lSpace do
      lNodeRA[lI+lS].rgba.rgbGreen := lNodeRA[lI].rgba.rgbGreen + round(lS * lSlope);
    //interpolate blue
    lSlope := (lNodeRA[lI+1].rgba.rgbBlue-lNodeRA[lI].rgba.rgbBlue)/lSpace;
    for lS := 1 to lSpace do
      lNodeRA[lI+lS].rgba.rgbBlue := lNodeRA[lI].rgba.rgbBlue + round(lS * lSlope);
    //interpolate alpha
    lSlope := (lNodeRA[lI+1].rgba.rgbreserved-lNodeRA[lI].rgba.rgbreserved)/lSpace;
    for lS := 1 to lSpace do
      lNodeRA[lI+lS].rgba.rgbreserved := lNodeRA[lI].rgba.rgbreserved + round(lS * lSlope);*)
  end;
end;

(*procedure GenerateLUT(lNodeRA: TCLUTnodeRA; lNodes: integer; var lCLUT: TCLUT);
var
  lSlope: single;
  lSpace,lI,lIprev,lS: integer;
  lMin,lMax: TCLUTnode;
begin
  for lI := 0 to 255 do begin
    lCLUT[lI] := RGBA(lI,lI,0,128);
  end;
end; *)

procedure LinearMinMaxCLUT(lMin,lMax: TGLRGBQuad; var lCLUT: TLUT);
//creates linear RGBA values from minimum to maximum
var
  lNodeRA: TCLUTrec;
begin
  lNodeRA.nodes[0].intensity := 0;
  lNodeRA.nodes[0].rgba := lMin;
  lNodeRA.nodes[1].intensity := 255;
  lNodeRA.nodes[1].rgba := lMax;
  lNodeRA.numnodes := 2;
  GenerateLUT(lNodeRA,lCLUT);
end;



procedure AutoContrast (var lCLUTrec: TCLUTrec);
begin
  lCLUTrec.nodes[0] := node(0,0,0,0,0);
  lCLUTrec.nodes[1] := node(255,255,255,255,100);
  lCLUTrec.numnodes := 2;
  lCLUTrec.min := 0;
  lCLUTrec.max := 0;
  //RangeRec(gTexture3D.MinThreshScaled,gTexture3D.MaxThreshScaled);
end;

function CLUT2disk(lRead: boolean; lFilename: string; var lCLUTrec: TCLUTrec): boolean;
//Read or write initialization variables to disk
var
  lIniFile: TIniFile;
  lI: integer;
begin
  result := false;
  if lRead then
    AutoContrast(lCLUTrec);
  if (lRead) and (not Fileexists(lFilename)) then
        exit;
  lIniFile := TIniFile.Create(lFilename);
  IniFloat(lRead,lIniFile, 'min',lCLUTrec.min);
  IniFloat(lRead,lIniFile, 'max',lCLUTrec.max);
  IniInt(lRead,lIniFile, 'numnodes',lCLUTrec.numnodes);
  if (lCLUTrec.numnodes > 1) and (lCLUTrec.numnodes <= 256) then begin
    for lI := 0 to (lCLUTrec.numnodes-1) do begin
      IniByte(lRead,lIniFile, 'nodeintensity'+inttostr(lI),lCLUTrec.nodes[lI].intensity);
      IniRGBA(lRead,lIniFile, 'nodergba'+inttostr(lI),lCLUTrec.nodes[lI].rgba);
    end;
  end else
    AutoContrast (lCLUTrec);
  lIniFile.Free;
  result := true;
end;

procedure CLUT2TLUT(lFilename: string; var lLUT: TLUT; var lCLUTrec: TCLUTrec);
begin
  if not CLUT2disk(true, lFilename,lCLUTrec) then
    exit;
  GenerateLUT(lCLUTrec, lLUT);
end;

procedure UpdateColorSchemes (var LUTdrop: TComboBox);
var
	lSearchRec: TSearchRec;
  lS: TStringList;
begin

  lS := TStringList.Create;
  if FindFirst(CLUTdir+pathdelim+'*.clut', faAnyFile, lSearchRec) = 0 then
	 repeat
      lS.Add(ParseFileName(ExtractFileName(lSearchRec.Name)))
		  //LUTdrop.Items.Add(ParseFileName(ExtractFileName(lSearchRec.Name)));
      //convertCLUT(CLUTdir+pathdelim+lSearchRec.Name);
	 until (FindNext(lSearchRec) <> 0);
  FindClose(lSearchRec);
  lS.sort;
  LUTdrop.Items.AddStrings(lS);
  Freeandnil(lS);
end;//UpdateColorSchemes

procedure SetItemNameX (lStr: string; var LUTdrop: TComboBox);
var
  i: integer;
  lIStr,lUStr,lPStr: string;
begin
  if LUTdrop.Items.Count < 1 then exit;
  lUStr := ansiuppercase(lStr);
  lPStr := parsefilename(extractfilename(lUStr));
  i := 0;
  while i < LUTdrop.Items.Count do begin
    lIStr := ansiuppercase(LUTdrop.Items[i]);
    if (lIStr = lUStr) or (lIStr = lPStr) then begin
      LUTdrop.ItemIndex := i;
      exit;
    end;
    inc(i);
  end;
end;//UpdateColorSchemes

end.
